function w_out = basicPhaseImethod(w_in, hs, Hs, SINRthresh, receiveNoisePower, ...
    noGroups, groupAssignments, noIterations)

% w_out = basicPhaseImethod(w_in, hs, Hs, SINRthresh, receiveNoisePower,
% noGroups, groupAssignments, noIterations)
%
% Assuming that for the given hs (i.e., the given phase rotated versions of
% the channel vectors), the SOCP approximation of the original problem is
% feasible, this functions attampts to find a strictly feasible point for
% this SOCP approximation using basic phase I method. Basic phase I method
% first increases the feasible set via relaxing the constraints by z to
% make w_in a strictly feasible point and then minimizes z.
%
% === Input: ===
% w_in:                 Beamforming weight vectors which is not yet
%                       feasible for the feasible SOCP approximation.
% hs:                   Matrix whose columns are the phase-rotated versions
%                       of the frequency-falt channel vectors for which the
%                       SOCP approximation is feasible.
% Hs:                   Array containing the matrices H_m for all the users
%                       for interference and noise in the SOC formulation.
% SINRthresh:           SINR threshold, minimum SINR that is guaranteed by
%                       the QoS constraints.
% receiveNoisePower:    Noise power at the receivers/users.
% noGroups:             Number of multicasting groups.
% groupAssignments:     Vector assigning the users to different multicast
%                       groups. The mth element is the index of the
%                       multicast group user m belongs to.
% noIterations:         Maximum number of iterations performed.
%
% === Output: ===
% w_out:                Beamforming weight vectors which are strictly
%                       feasible for the SOCP approximation if this method
%                       succeeded.
%
% Author: Nils Bornhorst


% number of transmit antennas:
noAntennas = length(hs(:, 1));

% number of receivers:
noReceivers = length(hs(1, :));


%% Some settings

% variable for counting number of iterations in phase I method to avoid
% infinite loop:
iterCount = 0;

% initialize the phase I method with w_in:
w = w_in;

% initialize z which indicates the maximum violation of the QoS
% constraints:

temp = zeros(noReceivers, 1);
for m = 1 : noReceivers
    % extract kth w (sub-)vector:
    w_k = w((groupAssignments(m) - 1)*noAntennas + 2 : groupAssignments(m)*noAntennas + 1);
    
    % calculate amount by which mth QoS constraint is satisfied or
    % violated:
    temp(m) = sqrt(SINRthresh)*norm(Hs(:,:,m)*w) - real(w_k'*hs(:,m));
end
% check whether w is already strictly feasible and break in that case:
if max(temp) < 0
    noIterations = 0;
end
% calculate a z which is a little larger than the largest violation:
z = max(temp) + 1;
    

%% Basic phase I method (to find a strictly feasible starting w):

% set initial barrier accuracy parameter:
t = 1;

% factor by which t is increased in every iteration:
mu = 8;

while (z >= 0) && (iterCount < noIterations)

    % solve the logarithmic barrier approximation of the SOCP using
    % Newton's method:
    [z, w] = newtonsMethodBasicPhaseI(z, w, hs, Hs, SINRthresh, receiveNoisePower, ...
        noGroups, groupAssignments, t);
    
    % increase barrier accuracy parameter
    t = mu*t;
    
    iterCount = iterCount + 1;   
    
end



%% Result:

w_out = w;

