function w_out = newtonsMethod(w_start, hs, Hs, SINRthresh, receiveNoisePower, ...
    noGroups, groupAssignments, t)

% w_out = newtonsMethod(w_start, hs, Hs, SINRthresh, receiveNoisePower,
% noGroups, groupAssignments, t)
%
% This function solves the logarithmic barrier approximation of the SOCP
% problem using Newton's method with backtracking line search.
%
% === Input: ===
% w_start:              Beamforming weight vectors as starting point for
%                       Newton's method.
% hs:                   Matrix whose columns are the phase-rotated versions
%                       of the frequency-falt channel vectors.
% Hs:                   Array containing the matrices H_m for all the users
%                       for interference and noise in the SOC formulation.
% SINRthresh:           SINR threshold, minimum SINR that is guaranteed by
%                       the QoS constraints.
% receiveNoisePower:    Noise power at the receivers/users.
% noGroups:             Number of multicasting groups.
% groupAssignments:     Vector assigning the users to different multicast
%                       groups. The mth element is the index of the
%                       multicast group user m belongs to.
% t:                    Barrier accuracy parameter. The larger t, the
%                       steeper the logarithmic barrier and thus, the more
%                       accurate the approximation.
%
% === Output: ===
% w_out:                Vector containing the beamforming weight vectors
%                       obtained by solving the approximation of the SOCP
%                       using Newton's method.
%
% Author: Nils Bornhorst


% number of transmit antennas:
noAntennas = length(hs(:, 1));

% number of receivers:
noReceivers = length(hs(1, :));

% just temporarily for stopping criterion:
lineSearchIteration = 1;
maxLineSearchIterations = 10;

% some parameters for Newton's method:

% for stopping criterion:
epsilon = 0.5;
% flattening factor in line search:
alpha = 0.01;
% step size reduction factor in line search:
beta = 0.25;


%% Conversion into real domain:

% New definitions of matrices and vectors follow which allow to convert the
% required functions from complex domain into real domain. The new real
% vectors are twice as long as the complex counterparts where the real and
% imaginary parts are stacked on top of each other:

Hs_real = zeros(2*noAntennas*noGroups, 2*noAntennas*noGroups, noReceivers);
for m = 1 : noReceivers
    % extract mth U matrix without the receiveNoisePower^2 and square it:
    H_m = Hs(2 : noAntennas*noGroups + 1, 2 : noAntennas*noGroups + 1, m)^2;
    
    Hs_real(:,:,m) = [real(H_m), -imag(H_m);...
                      imag(H_m), real(H_m)];
end
Hs = Hs_real;

hs_real = zeros(2*noAntennas*noGroups, noReceivers);
for m = 1 : noReceivers
    h_m = hs(:,m);
    for k = 1 : noGroups
        % only put the channel vector at the position of the desired group,
        % otherwise zeros:
        if k == groupAssignments(m)
            % put the real part at the respective position in the first
            % half of the real vector:
            hs_real((k - 1)*noAntennas + 1 : k*noAntennas, m) = real(h_m);
            % put the imaginary part at the respective position in the
            % second half of the real vector:
            hs_real(noAntennas*noGroups + (k - 1)*noAntennas + 1 : ...
                noAntennas*noGroups + k*noAntennas, m) = imag(h_m);
        end
    end
end
hs = hs_real;

% remove the first element of w_start which is = 1:
w_start = w_start(2 : noAntennas*noGroups + 1);
w = [real(w_start); imag(w_start)];


%% Newton's method:

% initialize the constraint function values:
constrFunctns = zeros(1, noReceivers);

% calculate the Hessians of the contraint functions (constant in all
% iterations):
hessians = zeros(2*noAntennas*noGroups, 2*noAntennas*noGroups, noReceivers);
for m = 1 : noReceivers
    hessians(:,:,m) = 2*hs(:,m)*hs(:,m)' - SINRthresh*(Hs(:,:,m) + Hs(:,:,m)');
end


for newtonIteration = 1 : 20
    
    
    %% Compute the Newton step and decrement:
    
    % calculate the part of the gradient corresponding to the objective
    % function:
    gradient = 2*t*w;
    
    % calculate the part of the hessian corresponding to the objective
    % function:
    hessian = 2*t*eye(2*noAntennas*noGroups);
    
    % now add the parts corresponding to the logarithmic barrier:
    for m = 1 : noReceivers       
        constrFunctns(m) = (w'*hs(:,m))^2 - SINRthresh*receiveNoisePower ...
            - SINRthresh*w'*Hs(:,:,m)*w;
        
        % calculate the gradient of the mth constraint function:
        gradient_temp = 2*(w'*hs(:,m))*hs(:,m) - SINRthresh*(Hs(:,:,m) + Hs(:,:,m)')*w;
        
        gradient = gradient - 1/constrFunctns(m) * gradient_temp;
        
        hessian = hessian + 1/constrFunctns(m)^2 * gradient_temp*gradient_temp'...
            - 1/constrFunctns(m) * hessians(m);
        
    end
    
    % diagonal loading in order to prevent ill-conditioned hessians:
    if rcond(hessian) < 0.1
        hessian = hessian + eye(2*noAntennas*noGroups);
    end
    
    newtonStep = -hessian^(-1)*gradient;
    
    newtonDecrement = -gradient'*newtonStep;
    
    
    %% Stopping criterion:
    
    % if stopping criterion for Newton's method satisfied
    % or if line search in previous iteration not successful -> change in w
    % negligible due to tiny step size -> makes no sense to continue:
    if newtonDecrement/2 <= epsilon || lineSearchIteration == maxLineSearchIterations
        break;
    end
    
    
    %% Backtracking line search:
    
    stepSize = 1;
    
    % evaluate function value at the current position:
    funcVal = t*w'*w - sum(log(constrFunctns));
    
    for lineSearchIteration = 1 : maxLineSearchIterations
        % calculate candidate for updated wheight vector:
        w_new = w + stepSize*newtonStep;
        
        % calculate function value corresponding to this new candidate:
        funcValNew = t*w_new'*w_new;
        for m = 1 : noReceivers
            constrFunctnsNew = (w_new'*hs(:,m))^2 ...
                - SINRthresh*receiveNoisePower - SINRthresh*w_new'*Hs(:,:,m)*w_new;
            
            funcValNew = funcValNew - log(constrFunctnsNew);
        end
        
        % check whether step size needs to be further decreased:
        
        % if we are outside of the domain (i.e., outside feasible set):
        if abs(imag(funcValNew)) > 0
            % reduce step size:
            stepSize = beta*stepSize;
        % if the new funtion value is still larger than the flattened
        % linearization in the direction of the Newton step:
        elseif funcValNew > funcVal + alpha*stepSize*gradient'*newtonStep
            % reduce step size:
            stepSize = beta*stepSize;
        else
            break;
        end           
    end
    
    
    %% Update:
    
    w = w + stepSize*newtonStep;
        
end


%% Conversion back to complex domain:

% initialize output weight vector:
w_out = zeros(noAntennas*noGroups + 1, 1);

% first element = 1 according to former SOCP formulation:
w_out(1) = 1;

% convert back to complex domain:
w_out(2 : noAntennas*noGroups + 1) = w(1 : noAntennas*noGroups) ...
    + 1i*w(noAntennas*noGroups + 1 : 2*noAntennas*noGroups);

